// DigitalIn.java

/*
This software is part of the Aplu FtController package.
It is Open Source Free Software, so you may
- run the code for any purpose
- study how the code works and adapt it to your needs
- integrate all or parts of the code in your own programs
- redistribute copies of the code
- improve the code and release your improvements to the public
However the use of the code is entirely your responsibility.
 */

package ch.aplu.ftcontroller;

/**
 * A class the represents the digital input ports (I1..I8).
 */
public class DigitalIn
{
  private FtController fc;
  private Object monitor = new Object();
  private volatile boolean isRelease = false;

  protected DigitalIn(FtController fc)
  {
    this.fc = fc;
  }

  /**
   * Polls the digital input ports (all 8 at the same time) and
   * returns their states as a bit pattern:
   * I1->bit 0 (LSB),...,I8->bit 7 (MSB), higher bits set to zero.
   * @return a bit pattern of the port states (integer 0..255)
   */
  public int getValue()
  {
    int rc;
    synchronized(fc)
    {
      rc = fc.nh.invoke(10);
    }
    return rc;
  }

  /**
   * Polls the digital input ports (all 8 at the same time) and
   * returns true if the specified port is turned on.
   * @return true, if the port is set to 1; false, if set to 0
   * @param portNb the port number 1..8
   */
  public boolean isPortOn(int portNb)
  {
    int bitValue = 1 << (portNb - 1);
    return ((getValue() & bitValue) != 0);
  }

  /**
   * Polls the digital input ports (all 8 at the same time) and
   * returns true if the specified port is turned off.
   * @return true, if the port is set to 0; false, if set to 1
   * @param portNb the port number 1..8
   */
  public boolean isPortOff(int portNb)
  {
    return !isPortOn(portNb);
  }

  private void waitPortChange(final int portNb, final boolean on)
  {
    if (isRelease)
      return;

    new Thread()
    {
      boolean isRunning = true;

      public void run()
      {
        while (isRunning)
        {
          if (on && isPortOn(portNb))
          {
            isRunning = false;
            wakeUp();
          }
          if (!on && !isPortOn(portNb))
          {
            isRunning = false;
            wakeUp();
          }
          fc.delay(1);
        }
      }
    }.start();
    putSleep();
  }

  /**
   * Wait until the specified port is set to 1.
   * @param portNb the port number 1..8
   */
  public void waitPortOn(int portNb)
  {
    waitPortChange(portNb, true);
  }

  protected void release()
  {
    isRelease = true;
    wakeUp();
  }

  /**
   * Wait until the specified port is set to 0.
   * @param portNb the port number 1..8
   */
  public void waitPortOff(int portNb)
  {
    waitPortChange(portNb, false);
  }

  private void putSleep()
  {
    synchronized (monitor)
    {
      try
      {
        monitor.wait();
      }
      catch (InterruptedException ex)
      {
      }
    }
  }

  private void wakeUp()
  {
    synchronized (monitor)
    {
      monitor.notifyAll();
    }
  }
}
