// Motor.java

/*
This software is part of the Aplu FtController package.
It is Open Source Free Software, so you may
- run the code for any purpose
- study how the code works and adapt it to your needs
- integrate all or parts of the code in your own programs
- redistribute copies of the code
- improve the code and release your improvements to the public
However the use of the code is entirely your responsibility.
 */

package ch.aplu.ftcontroller;

import javax.swing.*;
import java.awt.*;

/**
 * Class to handle the 4 motors at motor ports M1..M4.
 * For all methods with a step parameter, the motor is put in the so-called 
 * "RobMode". Two switches/counters are used: The "step switch" is preset to
 * the number of steps and counts left each time the switch 
 * is pressed/released. When the count reaches 0, the motor stops. When the
 * "end switch" is pressed the step switch is set to 0 and the motor 
 * stops immediately. The end switch is only active, if the motor turns left
 * (if motor.right(n) is called).
 * The following switch assignement is assumed:<br><br>
 * motor nb 1: end switch I1, step switch I2<br>
 * motor nb 2: end switch I3, step switch I4<br>
 * motor nb 3: end switch I5, step switch I6<br>
 * motor nb 4: end switch I7, step switch I8<br><br>
 *
 * In RobMode a registered counter listener at the step switch reports the 
 * count changes, but the counter at the end switch is disabled and a 
 * registered counter listener at the end switch will not report the press event.
 * It is recommended to register a MotorListener that reports each step change
 * and information whether the motor stopped because the step is reached or 
 * the end switch is pressed.
 */
public class Motor
{
  private FtController fc;
  private int motorNb;
  private boolean isRotating = false;
  private boolean isRobMode = false;
  private Object monitor = new Object();
  private boolean isEndSwitchPressed = false;

  protected Motor(FtController fc, int motorNb)
  {
    this.fc = fc;
    this.motorNb = motorNb;
  }

  /**
   * Returns number of motor (1..4).
   * @return the number of the motor
   */
  public int getNumber()
  {
    return motorNb;
  }

  /** 
   * Sets the rotation speed.
   * Keep in mind that the new speed is not modified 
   * until the next right/left method invocation.
   * Default value: 8.
   * @param speed the speed value 1..8 (1 is slowest)
   * @return a reference to the motor to allow chaining
   */
  public Motor setSpeed(int speed)
  {
    if (speed < 1 || speed > 8)
      return this;
    fc.speeds[motorNb - 1] = speed;
    return this;
  }

  /**
   * Stops the motor.
   * If the motor is not rotating, nothing happens.
   * @return a reference to the motor to allow chaining
   */
  public Motor stop()
  {
    if (!isRotating)
      return this;
    isRotating = false;
    isRobMode = false;
    fc.motorNb = motorNb;
    synchronized (fc)
    {
      fc.nh.invoke(2);
    }
    return this;
  }

  /** 
   * Starts turning the motor to the left (seen from front).
   * @return a reference to the motor to allow chaining
   */
  public Motor left()
  {
    isRotating = true;
    isRobMode = false;
    fc.motorNb = motorNb;
    synchronized (fc)
    {
      fc.nh.invoke(3);
    }
    return this;
  }

  /** 
   * Starts turning the motor to the right (seen from front).
   * @return a reference to the motor to allow chaining
   */
  public Motor right()
  {
    isRotating = true;
    isRobMode = false;
    fc.motorNb = motorNb;
    synchronized(fc)
    {
      fc.nh.invoke(4);
    }
    return this;
  }

  /**
   * Turns right given nb of steps to the left (seen from front).
   * After the motor is started, the method returns.
   * If the end switch is pressed, the motor stops immediately.
   * The motor is put in RobMode (see comment in the Motor class overview).
   * @param steps the number of switch actuations (press or release) of the
   * step switch
   * @return a reference to the motor to allow chaining
   */
  public Motor left(int steps)
  {
    fc.counterAry[2 * motorNb - 2].setCount(0);

    // Check of end switch is pressed
    // If true, notify event and return
    int din = fc.digitalIn.getValue();
    int mask = 1 << (2 * motorNb - 2);
    if ((din & mask) != 0)  // End switch pressed
    {
      if (fc.motorListeners[motorNb - 1] != null)
        fc.motorListeners[motorNb - 1].motorStopped(motorNb, true);
      isEndSwitchPressed = true;  // Report for left(steps, blocking)
      return this;
    }

    isRotating = true;
    isRobMode = true;
    fc.motorNb = motorNb;
    fc.steps[motorNb - 1] = steps;
    fc.enableEndSwitchReport[motorNb - 1] = true;
    synchronized (fc)
    {
      fc.nh.invoke(5);
    }
    return this;
  }

  /**
   * Turns right given nb of steps to the right (seen from front).
   * After the motor is started, the method returns.
   * The end switch is inactive.
   * The motor is put in RobMode (see comment in the Motor class overview).
   * @param steps the number of switch actuations (press or release) of the
   * step switch (see comment in the Motor class overview).
   * @return a reference to the motor to allow chaining
   */
  public Motor right(int steps)
  {
    fc.counterAry[2 * motorNb - 2].setCount(0);

    isRotating = true;
    isRobMode = true;
    fc.motorNb = motorNb;
    fc.steps[motorNb - 1] = steps;
    fc.enableEndSwitchReport[motorNb - 1] = false;  // Do not report
    synchronized(fc)
    {
     fc.nh.invoke(6);
    }
    return this;
  }

  /**
   * Same as left(steps) but if blocking = true, blocks until steps is reached
   * or the end switch is pressed. Warning: This method should never be used with
   * blocking = true from the Event Dispatch Thread (e.g. in a GUI control callback),
   * because the EDT will be blocked and the step counter no longer reports
   * when the number of steps are reached.
   * The callback methods of a registered MotorListener are still invoked.
   * @param steps the number of switch actuations (press or release)
   * @param blocking if true, the methods blocks until the step is reached
   */
  public Motor left(int steps, boolean blocking)
  {

    if (!blocking)
      return left(steps);

    if (EventQueue.isDispatchThread())
    {
        JOptionPane.showMessageDialog(null, "Blocking Event Dispatch Thread not allowed",
          "Error in calling Motor.left()",
          JOptionPane.ERROR_MESSAGE);
        fc.exit();
    }

    isEndSwitchPressed = false;

    fc.saveMotorListener(this);
    fc.addMotorListener(new MotorListener()
    {
      public void stepChanged(int motorNb, int step)
      {
        if (fc.saveMotorListeners[motorNb - 1] != null)
          fc.saveMotorListeners[motorNb - 1].stepChanged(motorNb, step);
      }

      public void motorStopped(int motorNb, boolean isEnd)
      {
        if (fc.saveMotorListeners[motorNb - 1] != null)
          fc.saveMotorListeners[motorNb - 1].motorStopped(motorNb, isEnd);
        wakeUp();
      }
    }, motorNb, false);
    left(steps);
    if (!isEndSwitchPressed)
      putSleep();
    fc.restoreMotorListener(this);
    return this;
  }

  /**
   * Same as right(steps) but if blocking = true, blocks until steps is reached.
   * The end switch is inactive. Warning: This method should never be used with
   * blocking = true  from the Event Dispatch Thread (e.g. in a GUI control callback),
   * because the EDT will be blocked and the step counter no longer reports
   * when the number of steps are reached.
   * The callback methods of a registered MotorListener are still invoked.
   * @param steps the number of switch actuations (press or release)
   * @param blocking if true, the methods blocks until the step is reached
   */ 
  public Motor right(int steps, boolean blocking)
  {
    if (!blocking)
      return right(steps);
    
    if (EventQueue.isDispatchThread())
    {
        JOptionPane.showMessageDialog(null, "Blocking Event Dispatch Thread not allowed",
          "Error in calling Motor.left()",
          JOptionPane.ERROR_MESSAGE);
        fc.exit();
    }

    fc.saveMotorListener(this);
    
    fc.addMotorListener(new MotorListener()
    {
      public void stepChanged(int motorNb, int step)
      {
        if (fc.saveMotorListeners[motorNb - 1] != null)
          fc.saveMotorListeners[motorNb - 1].stepChanged(motorNb, step);
      }

      public void motorStopped(int motorNb, boolean isEnd)
      {
        if (fc.saveMotorListeners[motorNb - 1] != null)
          fc.saveMotorListeners[motorNb - 1].motorStopped(motorNb, isEnd);
        wakeUp();
      }
    }, motorNb, false);
     
    right(steps);
    putSleep();
    fc.restoreMotorListener(this);
    return this;
  }

  /**
   * Returns state of motor.
   * @return true, if rotating, otherwise false
   */
  public boolean isRotating()
  {
    return isRotating;
  }

  protected void setRotating(boolean rotating)
  {
    isRotating = rotating;
  }

  protected boolean isRobMode()
  {
    return isRobMode;
  }

  private void putSleep()
  {
    synchronized (monitor)
    {
      try
      {
        monitor.wait();
      }
      catch (InterruptedException ex)
      {
      }
    }
  }

  private void wakeUp()
  {
    synchronized (monitor)
    {
      monitor.notifyAll();
    }
  }
}
